// SPDX-License-Identifier: GPL-2.0-only
/*
 *
 * Copyright (C) 2015 Naveen N. Rao, IBM Corporation
 */

#include "symbol.h" // for the elf__needs_adjust_symbols() prototype
#include <stdbool.h>

#ifdef HAVE_LIBELF_SUPPORT
#include <gelf.h>

bool elf__needs_adjust_symbols(GElf_Ehdr ehdr)
{
	return ehdr.e_type == ET_EXEC ||
	       ehdr.e_type == ET_REL ||
	       ehdr.e_type == ET_DYN;
}
#endif
